import json
from hestia_earth.schema import UNIQUENESS_FIELDS, Term, NODE_TYPES
from hestia_earth.schema.utils.sort import SORT_CONFIG
from flatten_json import flatten, unflatten_list
from collections import defaultdict
from copy import deepcopy

from ._shared import EXCLUDE_FIELDS, EXCLUDE_PRIVATE_FIELDS, _with_csv_formatting

pivot_exclude_fields = Term().fields
pivot_exclude_fields.update({k: "" for k in EXCLUDE_FIELDS} | {k: "" for k in EXCLUDE_PRIVATE_FIELDS})

term_exclude_fields = Term().fields
del term_exclude_fields["name"]
term_exclude_fields.update({k: "" for k in EXCLUDE_PRIVATE_FIELDS})

# Treat properties uniqueness fields as special case for now
PROPERTIES_VIRTUAL_UNIQUENESS_FIELD = "propertyValues"
ADAPTED_UNIQUENESS_FIELDS = deepcopy(UNIQUENESS_FIELDS)
for node_type, array_fields in UNIQUENESS_FIELDS.items():
    for array_field, uniqueness_fields in array_fields.items():
        if (
            "properties.term.@id" in uniqueness_fields
            and "properties.value" in uniqueness_fields
        ):
            uniqueness_fields.append(PROPERTIES_VIRTUAL_UNIQUENESS_FIELD)
            ADAPTED_UNIQUENESS_FIELDS[node_type][array_field] = [
                f
                for f in uniqueness_fields
                if f not in ("properties.term.@id", "properties.value")
            ]


def _do_pivot(node, parent_node_type=None, parent_field=None, level=0):  # noqa: C901
    # print('\ninput node', level, node, '\n')
    node_type = node.get("@type", node.get("type"))
    if node_type not in ADAPTED_UNIQUENESS_FIELDS:
        return node
    pivoted_node = {
        field: value
        for field, value in node.items()
        if all([
            field not in ADAPTED_UNIQUENESS_FIELDS[node_type],
            node_type != 'Term' or field not in term_exclude_fields
        ])
    }

    fields_to_pivot = [
        (field, uniqueness_fields)
        for field, uniqueness_fields in ADAPTED_UNIQUENESS_FIELDS[node_type].items()
        if field in node
    ]

    # print('\n', level, 'fields_to_pivot', fields_to_pivot)
    for field, uniqueness_fields in fields_to_pivot:
        # print('\nbefore processing node field', level, field, node[field], '\n')
        # Compress lists of 'Node' nodes to dict with single @id key.
        # The compressed field matches uniqueness fields like cycle.emissions.inputs.@id.
        if node[field] and SORT_CONFIG[node_type][field]["type"] in NODE_TYPES:
            pivoted_node[field] = {
                "@id": ";".join(
                    [
                        sub_node.get("@id", sub_node.get("id"))
                        for sub_node in node[field]
                    ]
                )
            }
            # print('\nafter processing node field', level, field, pivoted_node[field], '\n')
            continue
        else:
            node[field] = [
                _do_pivot(
                    term,
                    parent_node_type=node_type,
                    parent_field=field,
                    level=level + 1,
                )
                for term in node[field]
            ]
        # print('\nafter processing node field', level, field, node[field], '\n')

        pivoted_field = defaultdict(dict)
        # by_idx
        # [{
        #   term.@id: gwp100,
        #   any.term.field: value
        #   value: 10
        # }, etc.]
        by_idx = [flatten(term, ".") for term in node[field] if term]
        if not by_idx:
            continue
        # by_term_id
        # {
        #   gwp100: {
        #       indexes: [1, 4, etc.]
        #       combined_fields { field: last_traversed_value }
        #   }
        # }
        # print('\n\nby_idx', level, field, by_idx, '\n')
        by_term_id = defaultdict(lambda: {"indexes": [], "combined_fields": {}})
        # build list of property id:value pairs when required
        properties = []
        is_properties_special_case = (
            parent_node_type
            and field == "properties"
            and PROPERTIES_VIRTUAL_UNIQUENESS_FIELD
            in ADAPTED_UNIQUENESS_FIELDS[parent_node_type][parent_field]
        )
        for idx, term in enumerate(by_idx):
            id_key = "term.@id"
            id = term[id_key]
            if is_properties_special_case:
                properties.append(f'{id}[{term.get("value")}]')
            by_term_id[id]["indexes"].append(idx)
            combined_fields = {
                k: v
                for k, v in term.items()
                if k not in by_term_id[id]["combined_fields"]
            }
            by_term_id[id]["combined_fields"].update(combined_fields)
        # print('by_term_id', level, field, by_term_id, '\n')

        for term_id, term_data in by_term_id.items():
            indexes = term_data["indexes"]
            del term_data["combined_fields"][id_key]
            # print('combined_fields', field, term_id, term_data['combined_fields'], '\n')
            fields_to_include = {
                field: any(
                    by_idx[idx].get(field) != by_idx[indexes[0]].get(field)
                    for idx in indexes
                )
                for field in term_data["combined_fields"].keys()
                if field in uniqueness_fields
                or field != "value"
                and field.split(".")[-1] not in pivot_exclude_fields
            }
            # print('fields_to_include', level, field, term_id, fields_to_include, '\n')
            for idx in indexes:
                term = by_idx[idx]
                distingishing_field_fields = [
                    field
                    for field, not_unanimous in fields_to_include.items()
                    if field in uniqueness_fields
                    and (
                        # depthUpper and depthLower are exceptions which go into value col no matter what
                        field == "depthUpper"
                        or field == "depthLower"
                        or not_unanimous
                    )
                ]
                # print('distingishing_field_fields', level, field, term_id, distingishing_field_fields, '\n')
                unanimous_fields = {
                    field: term_data["combined_fields"][field]
                    for field, not_unanimous in fields_to_include.items()
                    if field not in distingishing_field_fields
                    and not not_unanimous
                    and field is not PROPERTIES_VIRTUAL_UNIQUENESS_FIELD
                }
                # print('unanimous_fields', level, field, term_id, unanimous_fields, '\n')
                differentiated_fields = {
                    field: term[field]
                    for field, not_unanimous in fields_to_include.items()
                    if field not in distingishing_field_fields
                    and not_unanimous
                    and field in term
                }
                # print('differentiated_fields', level, field, term_id, differentiated_fields, '\n')

                if unanimous_fields:
                    pivoted_field[term_id].update(unflatten_list(unanimous_fields, "."))
                distingishing_field_fields.sort()
                value_field_suffix = "".join(
                    [
                        f"+{term_field.split('.')[0]}[{term.get(term_field)}]"
                        for term_field in distingishing_field_fields
                        if term.get(term_field)
                    ]
                )
                distingishing_field = f"{term_id}{value_field_suffix}"
                if "value" in term:
                    pivoted_field[distingishing_field]["value"] = term.get("value")
                else:
                    pivoted_field[distingishing_field] = pivoted_field[
                        distingishing_field
                    ]
                if differentiated_fields:
                    pivoted_field[distingishing_field].update(
                        unflatten_list(differentiated_fields, ".")
                    )
        pivoted_node[field] = pivoted_field
        if is_properties_special_case:
            pivoted_node[PROPERTIES_VIRTUAL_UNIQUENESS_FIELD] = ";".join(properties)
    return pivoted_node


def pivot_json_node(json_node: str):
    """
    Pivot single schema-compliant unparsed json string node
    """
    node = json.loads(json_node, object_hook=_with_csv_formatting)
    return pivot_node(node)


def pivot_hestia_file(hestia_file: str):
    """
    Pivot json array of schema-compliant nodes on 'nodes' key of unparsed json string
    """
    parsed = json.loads(hestia_file, object_hook=_with_csv_formatting)
    return pivot_nodes(parsed.get('nodes', []))


def pivot_nodes(nodes: list[dict]):
    """
    Pivot multiple nodes in dict format parsed with object_hook=_with_csv_formatting
    """
    return [pivot_node(node) for node in nodes]


def pivot_node(node: dict):
    """
    Pivot single node in dict format parsed with object_hook=_with_csv_formatting
    """
    pivoted_node = _do_pivot(node)
    return pivoted_node



from functools import reduce
import numpy as np
from hestia_earth.schema import NodeType

# __package__ = "hestia_earth.utils" # required to run interactively in vscode
from .tools import flatten


PANDAS_IMPORT_ERROR_MSG = "Run `pip install pandas>=1.2` to use this functionality"
try:
    import pandas as pd

    version = [int(x) for x in pd.__version__.split(".")]
    if version[0] < 1 or (version[0] == 1 and version[1] < 2):
        raise ImportError(PANDAS_IMPORT_ERROR_MSG)
except ImportError:
    raise ImportError(PANDAS_IMPORT_ERROR_MSG)


def _replace_ids(df):
    # in columns, first letter is always lower case
    node_types = [e.value[0].lower() + e.value[1:] for e in NodeType]
    # add extra subvalues
    subvalues = ["source", "defaultSource", "site", "organisation", "cycle"]
    node_types = node_types + flatten(
        [v + "." + value for v in node_types] for value in subvalues
    )
    columns = reduce(
        lambda prev, curr: {**prev, curr + ".@id": curr + ".id"}, node_types, {}
    )
    return df.rename(columns=columns)


def _clean_term_columns(df):
    columns = ["name", "termType", "units"]
    cols = [c for c in df.columns if all([not c.endswith("." + v) for v in columns])]
    return df[cols]


def _replace_nan_values(df, col: str, columns: list):
    for index, row in df.iterrows():
        try:
            value = row[col]
            if np.isnan(value):
                for empty_col in columns:
                    df.loc[index, empty_col] = np.nan
        except TypeError:
            continue
    return df


def _empty_impact_na_values(df):
    impacts_columns = [c for c in df.columns if ".impacts."]
    impacts_values_columns = [c for c in impacts_columns if c.endswith(".value")]
    for col in impacts_values_columns:
        col_prefix = col.replace(".value", "")
        same_col = [c for c in impacts_columns if c.startswith(col_prefix) and c != col]
        _replace_nan_values(df, col, same_col)
    return df


def format_for_upload(filepath: str):
    """
    Format downloaded file for upload on HESTIA platform.
    Will replace all instances of `@id` to `id`, and drop the columns ending by `name`, `termType` or `units`.

    Parameters
    ----------
    filepath : str
        Path to the CSV to be formatted.

    Returns
    -------
    pandas.DataFrame
        Formatted pandas dataframe
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("Run `pip install pandas~=1.2.0` to use this functionality")

    df = pd.read_csv(filepath, index_col=None, na_values="")

    # replace @id with id for top-level Node
    df = _replace_ids(df)

    # drop all term columns that are not needed
    df = _clean_term_columns(df)

    # empty values for impacts which value are empty
    df = _empty_impact_na_values(df)

    return df



from functools import reduce
from io import StringIO
from typing import Union
import re
import requests
import csv
import numpy

from .storage import _load_from_storage
from .request import request_url, web_url

DELIMITER = '\t'
ENCODING = 'ISO-8859-1'
GLOSSARY_FOLDER = 'glossary/lookups'
_memory = {}
MISSING_VALUE = '-'
MISSING = -99999
_INDEX_COL = 'termid'


def _is_missing_value(value): return value == MISSING_VALUE or value == MISSING or value == str(MISSING)


def _replace_missing_values(value: str): return str(MISSING) if str(value) == '-' else value


def _rewrite_csv_file_as_tab(filepath: str):
    with open(filepath, 'r', encoding=ENCODING) as fp:
        reader = csv.reader(fp)
        for row in reader:
            yield DELIMITER.join(list(map(_replace_missing_values, row)))


def _rewrite_csv_text_as_tab(text: str):
    reader = csv.reader(StringIO(text))
    for row in reader:
        yield DELIMITER.join(list(map(_replace_missing_values, row)))


def _recfromcsv(data): return numpy.recfromcsv(data,
                                               missing_values=MISSING_VALUE,
                                               filling_values=MISSING,
                                               delimiter=DELIMITER,
                                               encoding=ENCODING)


def _memory_wrapper(key: str, func):
    global _memory  # noqa: F824
    _memory[key] = _memory[key] if key in _memory else func()
    return _memory[key]


def load_lookup(filepath: str, keep_in_memory: bool = False):
    """
    Import local lookup table as csv file into a `numpy.recarray`.

    Parameters
    ----------
    filepath : str
        The path of csv file on the local file system.
    keep_in_memory: bool
        Set to True if you want to store the file in memory for later use.

    Returns
    -------
    numpy.recarray
        The `numpy.recarray` converted from the csv content.
    """
    def load(): return _recfromcsv(_rewrite_csv_file_as_tab(filepath))
    return _memory_wrapper(filepath, load) if keep_in_memory else load()


def _download_lookup_data(filename: str):
    filepath = f"{GLOSSARY_FOLDER}/{filename}"

    def fallback():
        url = request_url(f"{web_url()}/{filepath}")
        return requests.get(url).content.decode('utf-8')

    try:
        data = _load_from_storage(filepath, glossary=True)
        return data.decode('utf-8') if data else None
    except ImportError:
        return fallback()


def _build_index(array: numpy.recarray):
    columns = list(array.dtype.names)
    return {
        row[_INDEX_COL]: {col: row[col] for col in columns}
        for row in array
    } if _INDEX_COL in columns else array


def download_lookup(filename: str, keep_in_memory: bool = True, build_index: bool = False):
    """
    Download lookup table from HESTIA as csv into a `numpy.recarray`.

    Parameters
    ----------
    filename : str
        The name on the file on the HESTIA lookup repository.
    keep_in_memory: bool
        Set to False if you do NOT want to store the file in memory for later use.
    build_index : bool
        Set to False to skip trying to build an index.

    Returns
    -------
    numpy.recarray
        The `numpy.recarray` converted from the csv content.
    """
    def load():
        data = _download_lookup_data(filename)
        rec = _recfromcsv(_rewrite_csv_text_as_tab(data)) if data else None
        return (_build_index(rec) if build_index else rec) if data else None

    try:
        return _memory_wrapper(filename, load) if keep_in_memory else load()
    except Exception:
        return None


def column_name(key: str):
    """
    Convert the column name to a usable key on a `numpy.recarray`.

    Parameters
    ----------
    key : str
        The column name.

    Returns
    -------
    str
        The column name that can be used in `get_table_value`.
    """
    return re.sub(r'[\(\)\,\.\-\'\\/"]', '', key).replace(' ', '_').lower() if key else ''


def _get_single_table_value(data: Union[dict, numpy.recarray], col_match: str, col_match_with, col_val):
    return (
        data.get(col_match_with, {})[col_val] if isinstance(data, dict) else
        data[data[col_match] == col_match_with][col_val][0]
    )


def _get_multiple_table_values(data: Union[dict, numpy.recarray], col_match: str, col_match_with, col_val):
    def reducer(x, values):
        col = values[1]
        value = col_match_with[values[0]]
        return x.get(value) if isinstance(x, dict) else x[x[col] == value]

    return reduce(reducer, enumerate(col_match), data)[col_val][0]


def get_table_value(lookup: Union[dict, numpy.recarray], col_match: str, col_match_with, col_val):
    """
    Get a value matched by one or more columns from a `numpy.recarray`.

    Parameters
    ----------
    lookup : dict | numpy.recarray
        The value returned by the `download_lookup` function.
    col_match : str
        Which `column` should be used to find data in. This will restrict the rows to search for.
        Can be a single `str` or a list of `str`. If a list is used, must be the same length as `col_match_with`.
    col_match_with
        Which column `value` should be used to find data in. This will restrict the rows to search for.
        Can be a single `str` or a list of `str`. If a list is used, must be the same length as `col_match`.
    col_val: str
        The column which contains the value to look for.

    Returns
    -------
    str
        The value found or `None` if no match.
    """
    single = isinstance(col_match, str) and isinstance(col_match_with, str)
    try:
        value = (
            _get_single_table_value(lookup, col_match, col_match_with, col_val) if single else
            _get_multiple_table_values(lookup, col_match, col_match_with, col_val)
        )
        return None if _is_missing_value(value) else value
    except Exception:
        return None


def find_term_ids_by(lookup: Union[dict, numpy.recarray], col_match: str, col_match_with):
    """
    Find `term.id` values where a column matches a specific value.

    Parameters
    ----------
    lookup : dict | numpy.recarray
        The value returned by the `download_lookup` function.
    col_match : str
        Which `column` should be used to find data in. This will restrict the rows to search for.
        Can be a single `str` or a list of `str`. If a list is used, must be the same length as `col_match_with`.
    col_match_with
        Which column `value` should be used to find data in. This will restrict the rows to search for.
        Can be a single `str` or a list of `str`. If a list is used, must be the same length as `col_match`.

    Returns
    -------
    list[str]
        The list of `term.id` that matched the expected column value.
    """
    term_ids = (
        set([
            key
            for key, value in lookup.items()
            if value.get(col_match) == col_match_with
        ])
    ) if isinstance(lookup, dict) else set(list(lookup[lookup[col_match] == col_match_with].termid))
    return list(map(str, term_ids))


def extract_grouped_data(data: str, key: str) -> str:
    """
    Extract value from a grouped data in a lookup table.

    Example:
    - with data: `Average_price_per_tonne:106950.5556;1991:-;1992:-`
    - get the value for `Average_price_per_tonne` = `106950.5556`

    Parameters
    ----------
    data
        The data to parse. Must be a string in the format `<key1>:<value>;<key2>:<value>`
    key
        The key to extract the data. If not present, `None` will be returned.

    Returns
    -------
    str
        The value found or `None` if no match.
    """
    grouped_data = reduce(lambda prev, curr: {
        **prev,
        **{curr.split(':')[0]: curr.split(':')[1]}
    }, data.split(';'), {}) if data is not None and isinstance(data, str) and len(data) > 1 else {}
    value = grouped_data.get(key)
    return None if _is_missing_value(value) else value


def extract_grouped_data_closest_date(data: str, year: int) -> str:
    """
    Extract date value from a grouped data in a lookup table.

    Example:
    - with data: `2000:-;2001:0.1;2002:0;2003:0;2004:0;2005:0`
    - get the value for `2001` = `0.1`

    Parameters
    ----------
    data
        The data to parse. Must be a string in the format `<key1>:<value>;<key2>:<value>`
    year
        The year to extract the data. If not present, the closest date data will be returned.

    Returns
    -------
    str
        The closest value found.
    """
    data_by_date = reduce(
        lambda prev, curr: {
            **prev,
            **{curr.split(':')[0]: curr.split(':')[1]}
        } if len(curr) > 0 and not _is_missing_value(curr.split(':')[1]) else prev,
        data.split(';'),
        {}
    ) if data is not None and isinstance(data, str) and len(data) > 1 else {}
    dist_years = list(data_by_date.keys())
    closest_year = min(dist_years, key=lambda x: abs(int(x) - year)) if len(dist_years) > 0 else None
    return None if closest_year is None else data_by_date.get(closest_year)


def lookup_term_ids(lookup: Union[dict, numpy.recarray]):
    """
    Get the `term.id` values from a lookup.

    Parameters
    ----------
    lookup : dict | numpy.recarray
        The value returned by the `download_lookup` function.

    Returns
    -------
    list[str]
        The `term.id` values from the lookup.
    """
    return lookup.keys() if isinstance(lookup, dict) else list(lookup.termid)


def lookup_columns(lookup: Union[dict, numpy.recarray]):
    """
    Get the columns from a lookup.

    Parameters
    ----------
    lookup : dict | numpy.recarray
        The value returned by the `download_lookup` function.

    Returns
    -------
    list[str]
        The columns from the lookup.
    """
    return list(list(lookup.values())[0].keys()) if isinstance(lookup, dict) else list(lookup.dtype.names)



from collections.abc import Iterable
from typing import Optional, Union, Any
from enum import Enum
from statistics import mode, mean

from .lookup import download_lookup, get_table_value, column_name
from .tools import non_empty_list


def get_lookup_value(blank_node: dict, column: str):
    term = blank_node.get('term', {})
    table_name = f"{term.get('termType')}.csv" if term else None
    value = get_table_value(
        download_lookup(table_name), 'termid', term.get('@id'), column_name(column)
    ) if table_name else None
    return value


class ArrayTreatment(Enum):
    """
    Enum representing different treatments for arrays of values.
    """
    MEAN = 'mean'
    MODE = 'mode'
    SUM = 'sum'
    FIRST = 'first'
    LAST = 'last'


def _should_run_array_treatment(value):
    return isinstance(value, Iterable) and len(value) > 0


DEFAULT_ARRAY_TREATMENT = ArrayTreatment.MEAN
ARRAY_TREATMENT_TO_REDUCER = {
    ArrayTreatment.MEAN: lambda value: mean(non_empty_list(value)) if _should_run_array_treatment(value) else None,
    ArrayTreatment.MODE: lambda value: mode(non_empty_list(value)) if _should_run_array_treatment(value) else None,
    ArrayTreatment.SUM: lambda value: sum(non_empty_list(value)) if _should_run_array_treatment(value) else None,
    ArrayTreatment.FIRST: lambda value: value[0] if _should_run_array_treatment(value) else None,
    ArrayTreatment.LAST: lambda value: value[-1] if _should_run_array_treatment(value) else None
}
"""
A dictionary mapping ArrayTreatment enums to corresponding reducer functions.
"""


def _retrieve_array_treatment(
    node: dict,
    is_larger_unit: bool = False,
    default: ArrayTreatment = ArrayTreatment.MEAN
) -> ArrayTreatment:
    """
    Retrieves the array treatment for a given node.

    Array treatments are used to reduce an array's list of values into
    a single value. The array treatment is retrieved from a lookup on
    the node's term.

    Parameters
    ----------
    node : dict
        The dictionary representing the node.
    is_larger_unit : bool, optional
        Flag indicating whether to use the larger unit lookup, by default `False`.
    default : ArrayTreatment, optional
        Default value to return if the lookup fails, by default `ArrayTreatment.MEAN`.

    Returns
    -------
    ArrayTreatment
        The retrieved array treatment.

    """
    ARRAY_TREATMENT_LOOKUPS = [
        'arrayTreatmentLargerUnitOfTime',
        'arrayTreatment'
    ]
    lookup = ARRAY_TREATMENT_LOOKUPS[0] if is_larger_unit else ARRAY_TREATMENT_LOOKUPS[1]

    lookup_value = get_lookup_value(node, lookup)

    return next(
        (treatment for treatment in ArrayTreatment if treatment.value == lookup_value),
        default
    )


def get_node_value(
    node: dict,
    key: str = 'value',
    is_larger_unit: bool = False,
    array_treatment: Optional[ArrayTreatment] = None,
    default: Any = 0
) -> Union[float, bool]:
    """
    Get the value from the dictionary representing the node,
    applying optional array treatment if the value is a list.

    Parameters
    ----------
    node : dict
        The dictionary representing the node.
    key : str
        The key to retrieve the value for. Will use `value` by default.
    is_larger_unit : bool, optional
        A flag indicating whether the unit of time is larger, by default `False`.
    array_treatment : ArrayTreatment, optional
        An optional override for the treatment to be applied to an array value, if `None` the array treatment in the
        node's term's lookup is used (which defaults to `FIRST` if no array treatment is specified), by default `None`.
    default : any
        The default value, if no value is found or it could not be parsed.

    Returns
    -------
    float | bool
        The extracted value from the node.
    """
    value = node.get(key)

    reducer = ARRAY_TREATMENT_TO_REDUCER[(
        array_treatment or _retrieve_array_treatment(
            node, is_larger_unit=is_larger_unit)
    )] if isinstance(value, list) and len(value) > 0 else None

    return reducer(value) if reducer else (
        value if any([isinstance(value, float), isinstance(value, int), isinstance(value, bool)])
        else (value or default)
    )
